/* ***********************************************************************
 * Diffpatch creates a patch routine to change a cbm file1 into cbm file2, 
 * appends the patch routine to file1 and saves it as a new cbm file. 
 *
 * September-2021 V0.1
 * Wilfried Elmenreich
 * Code under CC-BY-4.0 license
 * **************************************************** */

#define VERSION "0.1"

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>

#define verbose(...) if (verbose_flag) printf(__VA_ARGS__)

#define COND_PAL  1
#define COND_NTSC 2

#define ASM_PHA  0x48
#define ASM_PLA  0x68
#define ASM_LDAi 0xA9
#define ASM_LDA  0xAD
#define ASM_STA  0x8D
#define ASM_BNE  0xD0
#define ASM_BEQ  0xF0
#define ASM_JMP  0x4C
#define ASM_RTS  0x60

#define PAL_NTSC_FLAG 0x2A6 //on a Commodore 64

#define MAX_DIFFS 65536

#define lo(v) (v & 0xFF)
#define hi(v) (v>>8)

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
static int verbose_flag=0;
char *infilename1=NULL;
char *infilename2=NULL;
char *outfilename="out.prg";
int startoffset=-1;
int startaddr=-1;  
int patchstart=0;
int condition=0;

typedef struct { 
 uint16_t loadaddress;
 uint16_t length; 
 uint8_t data[0x10000];
} cbmfile_t;

typedef struct { 
 uint8_t value;
 uint16_t addr;
} diff_t;

char *progname;

cbmfile_t load_file(char* filename) {
  FILE *fp;
  cbmfile_t filedata;
  uint16_t nextaddress;
  
  fp = fopen(filename, "rb");
  if (fp==NULL) {
  	fprintf(stderr,"Error opening file %s!",filename);
  	exit(1);
  }
  
  filedata.loadaddress = fgetc(fp)+fgetc(fp)*256;
  nextaddress=filedata.loadaddress;
  
  filedata.length=0;
  while(1) {
    int c=fgetc(fp);
    if (feof(fp)) break;  
    filedata.data[nextaddress++]=c;
    filedata.length++;
  }
  fclose(fp);

  if (verbose_flag) {
      printf("Read file '%s' $%04x-$%04x\n",filename,filedata.loadaddress,nextaddress-1);
  }
  return filedata;
}
 
void save_file(cbmfile_t filedata, char* filename) {
  FILE *fp;
  uint16_t nextaddress;
   
  fp = fopen(filename, "wb");
  nextaddress=filedata.loadaddress;   
  fprintf(fp, "%c", nextaddress & 255);
  fprintf(fp, "%c", nextaddress / 256);
  for (int i=0; i<filedata.length; i++)
    fprintf(fp, "%c", filedata.data[nextaddress++]);
  fclose(fp);
  if (verbose_flag) {
    printf("Written file '%s' $%04x-$%04x\n",filename,filedata.loadaddress,nextaddress-1);
  }
}

cbmfile_t patch(cbmfile_t filedata1,cbmfile_t filedata2) {
  uint8_t usedbytes[256]= {0}; // If an array is partially initialized, elements that are not initialized receive the value 0 of the appropriate type
  diff_t diffs[MAX_DIFFS];
  int n_usedbytes=0;
  int n_diffs=0;
  uint16_t nextaddress;
  uint16_t patchcodeentry,patchcode_length,jmp_target;
  
  //compare the files
  for(int i=0;i<filedata1.length;i++) {
  	if (filedata1.data[filedata1.loadaddress+i] != filedata2.data[filedata2.loadaddress+i] ) {
  	  if (usedbytes[filedata2.data[filedata2.loadaddress+i]]==0) n_usedbytes++;
  	  usedbytes[filedata2.data[filedata2.loadaddress+i]]=1;
  	  diffs[n_diffs].addr  =filedata1.loadaddress+i;
  	  diffs[n_diffs].value =filedata2.data[filedata2.loadaddress+i];
  	  n_diffs++;
	}
  }
  patchcode_length=n_usedbytes*2 + n_diffs*3;
  
  if (n_diffs==0) {
  	fprintf(stderr,"Warning: both files are identical, no patchcode generated!\n");
  	return filedata1;
  }
  
  if (patchcode_length>1024)
    fprintf(stderr,"Warning: patch routine is %i bytes long\n",patchcode_length);
  else
    verbose("Patch routine needs to patch %i bytes with %i different values.\n",n_diffs,n_usedbytes);
  
  nextaddress=filedata1.loadaddress+filedata1.length;
  patchcodeentry=nextaddress;
  filedata1.data[nextaddress++] = ASM_PHA;
  
  //put condition code
  if (condition)
    switch(condition) {
      case COND_PAL:
      	if (patchcode_length<128) {
      		filedata1.data[nextaddress++] = ASM_LDA;
      		filedata1.data[nextaddress++] = lo(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = hi(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = ASM_BEQ;
      		filedata1.data[nextaddress++] = patchcode_length;
		} else {
      		filedata1.data[nextaddress++] = ASM_LDA;
      		filedata1.data[nextaddress++] = lo(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = hi(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = ASM_BNE;
      		filedata1.data[nextaddress++] = 3;
      		filedata1.data[nextaddress++] = ASM_JMP;
      		jmp_target=nextaddress+2+patchcode_length;      		
      		filedata1.data[nextaddress++] = lo(jmp_target);			
      		filedata1.data[nextaddress++] = hi(jmp_target);			
		}
		break;
      case COND_NTSC:
      	if (patchcode_length<128) {
      		filedata1.data[nextaddress++] = ASM_LDA;
      		filedata1.data[nextaddress++] = lo(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = hi(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = ASM_BNE;
      		filedata1.data[nextaddress++] = patchcode_length;
		} else {
      		filedata1.data[nextaddress++] = ASM_LDA;
      		filedata1.data[nextaddress++] = lo(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = hi(PAL_NTSC_FLAG);
      		filedata1.data[nextaddress++] = ASM_BEQ;
      		filedata1.data[nextaddress++] = 3;
      		filedata1.data[nextaddress++] = ASM_JMP;
      		jmp_target=nextaddress+2+patchcode_length;
      		filedata1.data[nextaddress++] = lo(jmp_target);			
      		filedata1.data[nextaddress++] = hi(jmp_target);			
		}
		break;		
	}
  //put patch code
  for(int i=0;i<256;i++) {
  	if (usedbytes[i]) {
  	  filedata1.data[nextaddress++] = ASM_LDAi;
  	  filedata1.data[nextaddress++] = i;
  	  for(int j=0;j<n_diffs;j++) {
  	  	if (diffs[j].value==i) {
  	      filedata1.data[nextaddress++] = ASM_STA;
  	      filedata1.data[nextaddress++] = lo(diffs[j].addr);
  	      filedata1.data[nextaddress++] = hi(diffs[j].addr);  	  		
		}
	  }
	}
  }
  filedata1.data[nextaddress++] = ASM_PLA;  //restore A
  
  if (patchstart) {
    if (filedata1.data[filedata1.loadaddress]!=ASM_JMP) {
      fprintf(stderr,"Cannot patch start because program does not start with a JMP.\n");
      exit(1);
	}	
    if (filedata2.data[filedata1.loadaddress]!=ASM_JMP) {
      fprintf(stderr,"Cannot patch start because program of patched file does not start with a JMP.\n");
      exit(1);
	}	
    if ((filedata1.data[filedata1.loadaddress+1]!=filedata2.data[filedata1.loadaddress+1])||(filedata1.data[filedata1.loadaddress+2]!=filedata2.data[filedata1.loadaddress+2])) {
      fprintf(stderr,"Cannot patch start because programs have different start commands.\n");
      exit(1);
	}  	
	filedata1.data[nextaddress++] = ASM_JMP;
	filedata1.data[nextaddress++] = filedata1.data[filedata1.loadaddress+1];
	filedata1.data[nextaddress++] = filedata1.data[filedata1.loadaddress+2];
	filedata1.data[filedata1.loadaddress+1] = lo(patchcodeentry);
	filedata1.data[filedata1.loadaddress+2] = hi(patchcodeentry);
	
  } else {
  	if (verbose_flag==0) printf("Patch code entry: $%4x\n",patchcodeentry);
  	filedata1.data[nextaddress++] = ASM_RTS;
  }
  
  filedata1.length=nextaddress-filedata1.loadaddress;
  
  verbose("Patch code $%4x-$%4x (%i bytes)\n",patchcodeentry,nextaddress,nextaddress-patchcodeentry);
  return filedata1;
}

void usage(char *s) {
   if (s!=NULL) fprintf(stderr,"%s\n\n",s);
   
   fprintf(stderr,"%s Version %s by Wil 2021\n",progname, VERSION);		
   fprintf(stderr,"\nUsage: %s [options] file1 file2\n",progname);
   fprintf(stderr,"\n%s creates a patch routine to change a cbm file1 into cbm file2, appends the patch routine to file1 and saves it as a new cbm file. A cbm file is a file for Commodore computers, where the first two bytes define the load address.\n",progname);     
   fprintf(stderr,"The current version is limited to single STA commands for each difference and the patchstart option only works if the code at the original start address starts with a JMP. Registers and most processor flags are preserved in the patch routine, but the Zero flag is changed by it.\n");     
   fprintf(stderr,"\nOptions:\n");  
   fprintf(stderr," -s,--startoffset  define the start offset relative to the load address\n");    
   fprintf(stderr," -a,--startaddress define the start address, must not be used together with start offset\n");       
   fprintf(stderr," -p,--patchstart   patch the start of the file so that the patch routine is executed first when the program is called via the start address\n");
   fprintf(stderr," -c,--condition    define a condition for the patch code to be executed. Possible values are PAL and NTSC\n");   
   fprintf(stderr," -o,--outfile      specify name of output file\n");      
   exit(1);
}


long strtoval(char *s) {
	int base=0;		//this allows numbers to be entered as C constants, that is decimals or using 0x or 0b as prefix for other number systems
	long value;
	char *eptr;
	
	if (*s=='$') {  //a number with a leading $ is interpreted as a hexadecimal number
		base=16;
		s++;
	}
	value=strtol(s, &eptr, base);
	if (eptr==s) {
		fprintf(stderr,"Cannot understand this number %s!\n",s);
		exit(1);
	}
	if (*eptr != '\0') {
		fprintf(stderr,"Trailing garbage on number %s!\n",s);
		exit(1);
	}	
	if (errno == ERANGE) {
		fprintf(stderr,"Number exceeds range %s!\n",s);
		exit(1);		
	}
	return value;
}

int main(int argc, char ** argv) {
  int c;
    
  progname=argv[0];

  while (1) { 
    static struct option long_options[] =
      {
        /* These options set a flag. */
        {"verbose", no_argument,       &verbose_flag, 1},
        /* These options dont set a flag.
           We distinguish them by their indices. */
        {"startoffset", required_argument, 0, 's'},
        {"startaddr",   required_argument, 0, 'a'},
        {"patchstart",  no_argument,       0, 'p'},
        {"condition",   required_argument, 0, 'c'},
        {"outfile",    required_argument,  0, 'o'},
        {0, 0, 0, 0}
      };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "s:a:pc:o:",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c)
      {
      case 0:
        /* If this option set a flag, do nothing else now. */
        if (long_options[option_index].flag != 0)
          break;
        printf ("option %s", long_options[option_index].name);
        if (optarg)
          printf (" with arg %s", optarg);
        printf ("\n");
        break;

      case 's':
        startoffset=strtoval(optarg);
        break;

      case 'a':
        startaddr=strtoval(optarg);
        break;

      case 'p':
        patchstart++;
        break;

      case 'c':
        if (!strcmp(optarg,"PAL")) condition=COND_PAL;
      	  if (!strcmp(optarg,"NTSC")) condition=COND_NTSC;          
      	  if (condition==0) {
      	  	fprintf(stderr,"Unknown condition %s\n",optarg);
      	  	exit(1);
      	  }
        break;

      case 'o':
        outfilename=optarg;
        break;

      case '?':
        /* getopt_long already printed an error message. */
        break;

      default:
        abort ();
      }
  }
  if ((startoffset>=0) && (startaddr>=0)) {
    fprintf(stderr,"Only one parameter, start offset or start address may be specified!\n");
    exit(1);
  }
  
  if (argc-optind!=2) usage(NULL);
  
  infilename1=argv[optind];
  infilename2=argv[optind+1];  
  
  cbmfile_t filedata1=load_file(infilename1);
  cbmfile_t filedata2=load_file(infilename2);  

  if (filedata1.loadaddress!=filedata2.loadaddress) {
    fprintf(stderr,"Warning: Load addresses of files are different!\n");
  }
  if (filedata1.length!=filedata2.length) {
    fprintf(stderr,"Warning: Lengths of files are different!\n");
  }  
  
  filedata1=patch(filedata1,filedata2);
  
  save_file(filedata1,outfilename);
  
  return 0;
}


